import random

TIMEOUT_SEC = 20.0


def get_user_agent() -> str:
    ua_list = [
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/114.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/104.0.5112.79 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/104.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/104.0.0.0 Safari/537.36",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/103.0.5060.53 Safari/537.36",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844.84 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.5112.79 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/121.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.5060.53 Safari/537.36",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.4844.84 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/117.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/116.0.5112.79 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/115.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/113.0.0.0 Safari/537.36",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/112.0.5060.53 Safari/537.36",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/111.0.4844.84 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/110.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.5112.79 Safari/537.36",
    ]
    return random.choice(ua_list)


def get_mobile_user_agent() -> str:
    ua_list = [
        "Mozilla/5.0 (iPhone; CPU iPhone OS 16_5 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.5 Mobile/15E148 Safari/604.1",
        "Mozilla/5.0 (iPad; CPU OS 16_5 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.5 Mobile/15E148 Safari/604.1",
        "Mozilla/5.0 (iPhone; CPU iPhone OS 16_5 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) CriOS/114.0.5735.99 Mobile/15E148 Safari/604.1",
        "Mozilla/5.0 (iPad; CPU OS 16_5 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) CriOS/114.0.5735.124 Mobile/15E148 Safari/604.1",
        "Mozilla/5.0 (Linux; Android 10; K) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/114.0.0.0 Mobile Safari/537.36",
        "Mozilla/5.0 (Linux; Android 13; SAMSUNG SM-S918B) AppleWebKit/537.36 (KHTML, like Gecko) SamsungBrowser/21.0 Chrome/110.0.5481.154 Mobile Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/113.0.0.0 Safari/537.36 OPR/99.0.0.0",
        "Mozilla/5.0 (Linux; Android 10; JNY-LX1; HMSCore 6.11.0.302) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844.88 HuaweiBrowser/13.0.5.303 Mobile Safari/537.36",
    ]
    return random.choice(ua_list)



from io import BytesIO
from typing import Optional

import magic
from playwright.async_api import async_playwright
from prometheus_client import Counter

from scraper.core.scrape_utils import TIMEOUT_SEC, get_user_agent

PLAYWRIGHT_COUNTER = Counter("playwright_request", "Number of api requests")

PLAYWRIGHT_FAIL_COUNTER = Counter(
    "playwright_fail_request", "Number of get requests that failed"
)


async def aget_playwright(url: str) -> Optional[str]:
    PLAYWRIGHT_COUNTER.inc()

    # https://github.com/Josh-XT/AGiXT/blob/main/agixt/extensions/web_playwright.py
    async with async_playwright() as p:
        browser = await p.chromium.launch(headless=True)
        context = await browser.new_context(
            viewport={"width": 1920, "height": 1080}, user_agent=get_user_agent()
        )
        await context.add_init_script(
            path="/envision/ml/scraper/core/libs/stealth.min.js"
        )
        page = await context.new_page()
        text = None
        try:
            response = await page.goto(url, timeout=TIMEOUT_SEC * 1_000)
            text = await response.body()
            content_type = magic.from_buffer(BytesIO(text).read(2048), mime=True)

            if "image" not in content_type:
                # Navigate to the URL with an increased timeout for JavaScript-heavy pages
                # Wait for potential JavaScript executions. Adjust or remove based on the specific needs of the target website
                await page.wait_for_timeout(2_000)  # Wait for an additional 2 seconds
                # Retrieve the page content after JavaScript execution
                text = await page.content()
        except TimeoutError as e:
            print(f"Page load timed out: {e}")
        except Exception as e:
            print(f"An error occurred: {e}")
        finally:
            await browser.close()

        if text is None:
            PLAYWRIGHT_FAIL_COUNTER.inc()

        return text



from pydantic import BaseModel
from typing import Optional, List, Union


class WebContentType:
    HTML = 1
    PDF = 2
    IMAGE = 3


class WebScrapeContent(BaseModel):
    content_type: int
    title: Optional[str] = None
    description: Optional[str] = None
    url: Optional[str] = None
    content: Optional[Union[str, bytes]] = None
    images: Optional[List[str]] = None
    texts: Optional[List[str]] = None
    thumbnail: Optional[str] = None
    thumbnail_content: Optional[Union[str, bytes]] = None
    thumbnail_width: Optional[int] = None
    thumbnail_height: Optional[int] = None


# class HtmlWebContent(WebScrapeContent):
#     link: Optional[str]
#     html: Optional[str]
#
#
# class ImageWebContent(WebScrapeContent):
#     image_url: str
#
#
# class PDFWebContent(WebScrapeContent):
#     link: str



import logging
from io import BytesIO

from typing import Optional, Tuple

from httpx import AsyncClient, Client
from PIL import Image


async def async_http_get_image(
    http_path: str,
    timeout_secs: Optional[int] = None
) -> Optional[Tuple[str, Image.Image]]:
    try:
        async with AsyncClient() as client:
            img_data = (await client.get(http_path, timeout=timeout_secs)).content
        return http_path, Image.open(BytesIO(img_data))
    except Exception as e:
        logging.error(e)
    return None



import modal
import asyncio

from typing import Optional

playwright_web_scraper = modal.Function.lookup("web-scraper", "playwright_web_scrape")


async def playwright_web_scrape(
    url: str, timeout: int = 60 * 5
) -> Optional[str]:
    try:
        result = await asyncio.wait_for(
            playwright_web_scraper.remote.aio(url), timeout=timeout
        )
        return result
    except asyncio.TimeoutError as e:
        print("Timeout web scraping hit")
        return None



from setuptools import setup, find_packages


setup(
    name='aideate_blade',
    version='1.0.1',
    description='web content scraper',
    author='Aideate',
    author_email='xiaye@aideate.ai',
    packages=find_packages(),
    install_requires=[
        # List your dependencies here
        "pydantic==2.6.1",
        "ipdb==0.13.11",
        "simplejson==3.17.6",
        "boto3==1.29.6",
        "boto3[crt]==1.29.6",
        "sqlalchemy==2.0.5.post1",
        "psycopg2==2.9.5",
        "scrapetube==2.3.1",
        "overrides==7.3.1",
        "snakemd==2.1.0",
        "html2text==2020.1.16",
        "playwright==1.35.0",
        "prometheus-client==0.20.0",
        "extraction==0.3",
        "pdfplumber==0.11.0",
        "stripe==5.5.0",
        "python-magic==0.4.27",
        "StrEnum==0.4.15",
        "modal==0.61.105",
        "setuptools==69.5.1",
    ],
)


import asyncio
import logging
import re
from enum import Enum
from io import BytesIO, StringIO
from typing import Optional, Union

import aiohttp
import extraction
import html2text
import magic
import pdfplumber
from PIL import Image
from prometheus_client import Counter
from snakemd import Document as mdDocument
from tenacity import retry, retry_if_exception_type, stop_after_attempt, wait_fixed

from scraper.core.image_utils import async_http_get_image
from scraper.core.schemas import WebScrapeContent, WebContentType
from scraper.modal.modal_client import playwright_web_scrape
from scraper.core.scrape_utils import TIMEOUT_SEC, get_user_agent
from settings import ZENROWS_API_KEY

logger = logging.getLogger(__name__)

LLM_SUMMARIZE_FAIL_COUNTER = Counter(
    "scrape_llm_summarize_fail_request", "Number of llm summarize requests that failed"
)

GET_COUNTER = Counter("scrape_get_request", "Number of get requests")
SCRAPING_API_COUNTER = Counter("scrape_api_request", "Number of api requests")

GET_FAIL_COUNTER = Counter(
    "scrape_get_fail_request", "Number of get requests that failed"
)
SCRAPING_API_FAIL_COUNTER = Counter(
    "scrape_api_fail_request", "Number of api requests that failed"
)


HEURISTIC_VALID_HTML_TEXT = 1000
MAX_FILE_SIZE = 512 * 1024 * 1024  # 512 mb




async def parse_html(
    html_text: str,
    url: str,
    image_url: Optional[str] = None,
) -> WebScrapeContent:
    """
    Extract the HTML into various structured metadata
    """
    web_content = WebScrapeContent(content_type=WebContentType.HTML, content=html_text)
    extracted = extraction.Extractor().extract(html_text, source_url=url)

    title, thumbnails, canonical_url = (
        extracted.title,
        extracted.images,
        extracted.url,
    )
    thumbnails = [
        thumbnail_url
        for thumbnail_url in thumbnails
        if thumbnail_url.startswith("http")
    ]
    web_content.url = url
    if canonical_url:
        web_content.url = canonical_url
    if title:
        web_content.title = title
    if image_url:
        try:
            _, image = await async_http_get_image(image_url, 1)
            web_content.thumbnail = image_url
            web_content.thumbnail_width = image.width
            web_content.thumbnail_height = image.height
        except Exception as e:
            logger.info(str(e))
    if not web_content.thumbnail and thumbnails:
        thumbnail = None
        max_width, max_height = 0, 0
        get_image_tasks = [
            async_http_get_image(thumbnail_url, 1)
            for thumbnail_url in thumbnails
        ]
        image_results = await asyncio.gather(*get_image_tasks)
        image_results = [r for r in image_results if r]
        for url, image in image_results:
            try:
                if image.width * image.height > max_width * max_height:
                    thumbnail = url
                    max_width, max_height = image.width, image.height
            except Exception as e:
                logger.info(str(e))
        if thumbnail:
            web_content.thumbnail = thumbnail
            web_content.thumbnail_width = max_width
            web_content.thumbnail_height = max_height

    if extracted.images:
        web_content.images = extracted.images

    return web_content


def parse_image(content_bytes: bytes, image_url: str) -> WebScrapeContent:
    web_content = WebScrapeContent(
        content_type=WebContentType.IMAGE,
        content=content_bytes,
        url=image_url,
        images=[image_url],
        thumbnail=image_url,
    )
    image = Image.open(BytesIO(content_bytes))
    web_content.thumbnail_width=image.width
    web_content.thumbnail_height=image.height
    return web_content


async def parse_pdf(content_bytes: bytes, url: str):
    web_content = WebScrapeContent(content_type=WebContentType.PDF, content=content_bytes)
    pdf_file = BytesIO(content_bytes)

    texts = []
    with pdfplumber.open(pdf_file) as pdf:
        if len(pdf.pages) > 0:
            text = pdf.pages[0].extract_text()
            if text:
                texts.append(text)
        thumbnail_bytes = BytesIO()
        im = pdf.pages[0].to_image(resolution=150, antialias=True)
        im.save(thumbnail_bytes, format="PNG")
        thumbnail_bytes.seek(0)

        web_content.thumbnail_content = thumbnail_bytes.getvalue()

    web_content.title = url
    web_content.texts = texts



class LinkType(str, Enum):
    # be careful of JSON serialization bugs: https://stackoverflow.com/questions/65339635/how-to-deserialise-enumeration-with-string-representation
    # not a scenario now

    HTML = "html"
    PDF = "pdf"
    IMAGE = "image"
    UNKNOWN = "unknown"


def _get_web_type(content: Union[str, bytes]) -> LinkType:
    if isinstance(content, bytes):
        file_buffer = BytesIO(content)
    else:
        file_buffer = StringIO(content)
    content_type = magic.from_buffer(file_buffer.read(2048), mime=True)
    if "text/html" in content_type:
        return LinkType.HTML
    elif "application/pdf" in content_type:
        return LinkType.PDF
    elif any(
        image_type in content_type
        for image_type in [
            "image/png",
            "image/jpeg",
            "image/webp",
            "image/gif",
            "image/svg+xml",
        ]
    ):
        return LinkType.IMAGE
    return LinkType.UNKNOWN


def extract_text_from_content(content: Union[str, bytes]) -> Optional[str]:
    link_type = _get_web_type(content)
    match link_type:
        case LinkType.HTML:
            if isinstance(content, bytes):
                content = content.decode("utf-8")
            return html2text.html2text(content)

        case LinkType.PDF:
            if isinstance(content, str):
                content = content.encode("utf-8")
            source_doc = mdDocument()
            with pdfplumber.open(BytesIO(content)) as pdf:
                for page in pdf.pages:
                    text = page.extract_text()
                    source_doc.add_paragraph(text)
                    tables = page.extract_tables()
                    for table in tables:
                        if table:
                            source_doc.add_table(header=table[0], data=table[1:])

            return str(source_doc)
        case LinkType.IMAGE:
            return None


async def process_web_content(
    content: Optional[Union[str, bytes]],
    url: str,
) -> Optional[WebScrapeContent]:
    if not content:
        return None

    link_type = _get_web_type(content)
    match link_type:
        case LinkType.HTML:
            return await parse_html(content, url)
        case LinkType.PDF:
            return await parse_pdf(content, url)
        case LinkType.IMAGE:
            return parse_image(content, url)
    return None


def is_instagram_url(url: str) -> bool:
    return "instagram.com" in url


async def ascrape_web(
    url: str
) -> Optional[WebScrapeContent]:
    if not is_instagram_url(url):
        # instagram doesn't work for these options but in general we want them first

        content = await aget_scrape_api(url)
        response = await process_web_content(content, url)
        if response:
            return response

        # fallback to requests
        content = await aget(url)
        response = await process_web_content(content, url)
        if response:
            return response

    content = await playwright_web_scrape(url, timeout=int(TIMEOUT_SEC))
    response = await process_web_content(content, url)
    if response:
        return response

    return None


class ContentExceedMaxLimit(Exception):
    pass


def make_default_headers(user_agent: str) -> dict:
    return {
        "accept": "*/*",
        "content-type": "application/json; charset=utf-8",
        "user-agent": user_agent,
        "accept-language": "en-us",
    }


async def aget(url: str) -> Optional[Union[str, bytes]]:
    GET_COUNTER.inc()
    # Send a GET request to the URL
    try:
        print("fetching url via GET", url)
        async with aiohttp.ClientSession() as session:
            async with session.get(
                url, headers=make_default_headers(get_user_agent()), timeout=TIMEOUT_SEC
            ) as response:
                print("received url")
                # If the GET request is successful, the status code will be 200
                if response.status == 200:
                    total_size = 0
                    bytes_io = BytesIO()
                    async for chunk in response.content.iter_chunked(n=1024 * 1024):
                        total_size += len(chunk)
                        if total_size > MAX_FILE_SIZE:
                            raise ContentExceedMaxLimit(
                                f"Uh oh, website content is larger than maximum support size of {MAX_FILE_SIZE} bytes"
                            )
                        bytes_io.write(chunk)
                    bytes_data = bytes_io.getvalue()
                    response._body = bytes_data
                    encoding = response.get_encoding()
                    try:
                        return str(bytes_data, encoding, errors="strict")
                    except UnicodeDecodeError as e:
                        return bytes_data

                GET_FAIL_COUNTER.inc()
                return None
    except ContentExceedMaxLimit as e:
        GET_FAIL_COUNTER.inc()
        raise e
    except Exception as e:
        GET_FAIL_COUNTER.inc()
        logger.error(f"cannot crawl url {url}", e)
        return None


def need_enable_javascript(html_text):
    # Check for the presence of <noscript> tags
    if re.search(r"<noscript>", html_text, re.IGNORECASE):
        return True

    # Check for JavaScript-dependent attributes
    if re.search(r"onclick|onload|onsubmit", html_text, re.IGNORECASE):
        return True

    # Check for variations of "JavaScript is disabled" text
    javascript_disabled_patterns = [
        r"javascript\s+is\s+disabled",
        r"enable\s+javascript",
        r"javascript\s+not\s+supported",
        r"please\s+enable\s+javascript",
        r"javascript\s+required",
        # Add more patterns as needed
    ]

    for pattern in javascript_disabled_patterns:
        if re.search(pattern, html_text, re.IGNORECASE):
            return True

    return False


@retry(
    stop=stop_after_attempt(3),
    wait=wait_fixed(2),
    retry=retry_if_exception_type(asyncio.exceptions.TimeoutError),
)
async def aget_scrape_api(url: str, enable_javascript: bool = False):
    SCRAPING_API_COUNTER.inc()
    params = {
        "apikey": ZENROWS_API_KEY,
        "url": url,
    }
    if enable_javascript:
        params["js_render"] = "true"

    async with aiohttp.ClientSession() as session:
        async with session.get(
            url="https://api.zenrows.com/v1/",
            params=params,
            timeout=TIMEOUT_SEC,
        ) as response:
            # If the GET request is successful, the status code will be 200
            if response.status != 200:
                SCRAPING_API_FAIL_COUNTER.inc()
                return None

            text = await response.text(errors="replace")
            if need_enable_javascript(text) and not enable_javascript:
                logger.info("enable javascript retry")
                return await aget_scrape_api(url, enable_javascript=True)
            return text


if __name__ == "__main__":
    import asyncio

    # print(asyncio.run(aget_scrapingbee("https://openai.com")))
    # result = asyncio.run(aget("https://arxiv.org/pdf/2307.01952.pdf"))
    # result = asyncio.run(aget_scrapingbee("https://arxiv.org/pdf/2307.01952.pdf"))
    # result = get("https://arxiv.org/pdf/2307.01952.pdf")
    # result = get_scrapingbee("https://arxiv.org/pdf/2307.01952.pdf")
    # result = scrape_web("https://arxiv.org/pdf/2307.01952.pdf")
    # result = ascrape_web("https://arxiv.org/pdf/2307.01952.pdf")
    # print(scrape_web("https://www.youtube.com/watch?v=SKia5QUiGkE"))
    # print(scrape_web("https://www.youtube.com/watch?v=SKia5QUiGkE"))
    # message = asyncio.run(
    #    ascrape_web("https://twitter.com/itsandrewgao/status/1786591700410118295")
    # )
    result = asyncio.run(ascrape_web("https://m-arkitektur.se/villa-hagerman-gotland"))
    print(result)



from typing import List, Optional

from modal import Function, Image, Mount, Secret, Stub, asgi_app, gpu, method

import proto.gen.message_pb2 as message_pb2
from modal_functions.tasks.base import image

workers_stub = Stub("web-scraper", image=image)


@workers_stub.function(secrets=[Secret.from_name("my-aws-secret")])
async def playwright_web_scrape(url) -> Optional[message_pb2.DocumentMessage]:
    from scraper.core.scrape_playwright import aget_playwright

    result = await aget_playwright(url)
    return result


# Test each method
@workers_stub.local_entrypoint()
async def test_methods():
    print(
        playwright_web_scrape.remote(url="https://www.youtube.com/shorts/o0Q6mK-hSIw")
    )
